package Ska::Message;

# use 5.008003;
use strict;
use warnings;
use FileHandle;
use Carp;
use POSIX qw(strftime);

require Exporter;

our @ISA = qw(Exporter);

# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

# This allows declaration	use Ska::Message ':all';
# If you do not need this, moving things directly into @EXPORT or @EXPORT_OK
# will save memory.
our %EXPORT_TAGS = ( 'all' => [ qw() ] );

our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );

our @EXPORT = qw();

our $VERSION = '0.01';

our %DEFAULT = (options => {stdout => 1,
			    auto_cr => 1,
			   },
		time => '<%Y-%b-%d %H:%M:%S> ',
		caller => '[%22.22s %4.4s] ');
our @OPTIONS = qw(stdout stderr file append time caller debug auto_cr);
			

##****************************************************************************
sub new {
##****************************************************************************
    my $classname = shift;
    my $self = {};
    bless ($self);
    local $_;

    # Set message parameters.  Only allow options in specified list
    %{$self} = (%{$DEFAULT{options}},
		@_
	       );

    # Open file handle if needed
    if (defined $self->{file} and $self->{file} =~ /\S+/) {
	my $append_mode = $self->{append} ? ">>" : ">";
	croak "ERROR - Could not open message file $self->{file}\n"
	  unless ($self->{file_handle} = new FileHandle "$append_mode $self->{file}");
	$self->{file_handle}->autoflush(1);
    }
    
    for (qw(time caller)) {
	$self->{$_} = $DEFAULT{$_} if (defined $self->{$_} and $self->{$_} !~ /%/);
    }

    return $self;
}

##****************************************************************************
sub message {
##****************************************************************************
    my $self = shift;
    local $_;

    # Set message options
    my %opt = map { $_ => $self->{$_} } @OPTIONS;

    # If there is a hash ref at end for extra options, then use these to override
    %opt = (%opt, %{pop @_}) if (ref $_[-1] eq 'HASH');
    for (qw(time caller)) {
	$opt{$_} = $DEFAULT{$_} if (defined $opt{$_} and $opt{$_} !~ /%/);
    }

    my $message = @_ ? join('', @_) : '';

    # Coming from our debug() routine?  If so, add one to the caller arg
    # below to get back to the real originator of the message request
    my $dbg = (caller(1) and (caller(1))[3] eq 'Ska::Message::debug') ? 1 : 0;

    my $time_str = $opt{time} ? strftime($opt{time}, localtime) : '';

    my $caller_str = '';
    if ($opt{caller}) {
	my $caller = (caller(1+$dbg))[0];
	my $line = (caller(0+$dbg))[2];
	$caller ||= $0;
	if (length $caller > 22) {
	    $caller = substr($caller,0,10)."..".substr($caller,-10,10);
	}
	$caller_str = sprintf($opt{caller}, $caller, $line);
    }

    my $align = ' ' x length("${time_str}${caller_str}");
    # Remove any trailing CR if auto_cr, then put in alignment space for
    # any interior CR, and finally put the CR back if needed
    $message =~ s/\n\Z// if $opt{auto_cr};
    $message =~ s/\n(.)/\n$align$1/g;
    $message .= "\n" if $opt{auto_cr};

    my $out = "${time_str}${caller_str}$message";
    my $fh = $self->{file_handle};
    print $fh $out if ($opt{file} and $self->{file_handle});
    print STDOUT $out if $opt{stdout};
    print STDERR $out if $opt{stderr};
}

##****************************************************************************
sub debug {
##****************************************************************************
    my $self = shift;
    my $level = (@_ and $_[0] =~ /\A \s* \d+ \s* \Z/x) ? shift : undef;
    
    if (defined $self->{debug} and (not defined $level or $self->{debug} >= $level)) {
	$self->message(@_, {stdout=>0,
			    stderr=>1,
			    time=>1,
			    caller=>1,
			    auto_cr=>1,
			   }
		      );
    }
}

# Preloaded methods go here.

1;
__END__
# Below is stub documentation for your module. You'd better edit it!

=head1 NAME

Ska::Message - Output messages in a nice way

=head1 SYNOPSIS

  use Ska::Message;

  $log = new Ska::Message (stdout => 1,
                           file   => 'task.log',
                           append => 1,
                           time   => 1,
                           debug  => $DEBUG);

  $log->message("I wants the ring, Frodo");
  $log->message("This message goes only to STDERR with no time",
                {stderr=>1, stdout=>0, time=>0, file=>undef});
  $log->debug(2, "Debug message if $DEBUG >= 2");
  $log->debug("Debug message for any defined value of $DEBUG");

=head1 METHODS

=over 8

=item B<new>

Create a new message object with specified options:

=over 8

=item B<stdout>

Send message to STDOUT.  Default = 1

=item B<stderr>

Send message to STDERR.  Default = undef

=item B<file>

Send message to specified file if supplied.  Default = undef

=item B<append>

Append to an existing file instead of clobbering it.  Default = undef

=item B<time>

Prepend the time to the message using the given strftime() format.  
If any value not containing '%' is given then '<%Y-%b-%d %H:%M> ' is used.
Default = undef

=item B<caller>

Print name and line number of the calling program using the given sprintf() format.
If any value not containing '%' is given then '[%22.22s %4.4s] ' is used.  Default = undef

=item B<debug>

Debug level for the message object that is used in the debug() method.  No debug messages 
will be output if this is undefined.  Default = undef

=item B<auto_cr>

Supply a carriage return at the end of message.  Default = 1.

=back

=item B<message>(<message>, ... [{opt => val, ...}

Print the supplied string(s) as specified in the object creation.  Any of those
original options can be overridden for this instance by supplying
a hash reference as the last argument.  The exception is that the
name of the output B<file> cannot be changed, although if B<file> is
undefined the message will not be logged to the file.

=item B<debug>(<debug_level>, <message>, ...)

If the object <debug> >= <debug_level> then do 
  message(<message>,..., {stdout => 0, 
                          stderr => 1, 
                          time   => 1, 
                          caller => 1,
                          auto_cr=> 1})

If the first argument doesn't look like a number (\d+) then it is
taken to be a message argument and <debug_level> is set to 0.

=back

=head1 DESCRIPTION

Log messages from program to a configurable set of output streams.

=head2 EXPORT

None by default.

=head1 AUTHOR

Tom Aldcroft, E<lt>aldcroft@localdomainE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2005 by Tom Aldcroft

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.3 or,
at your option, any later version of Perl 5 you may have available.

=cut
