# --8<--8<--8<--8<--
#
# Copyright (C) 2006 Smithsonian Astrophysical Observatory
#
# This file is part of yaxx
#
# yaxx is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# yaxx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the 
#       Free Software Foundation, Inc. 
#       51 Franklin Street, Fifth Floor
#       Boston, MA  02110-1301, USA
#
# -->8-->8-->8-->8--

package Yaxx;
use strict;
use warnings;

##****************************************************************************
sub make_dirs {
##****************************************************************************
    my $self = shift;
    my $opt = shift;

    foreach my $dir (listify($opt->{mkdir})) {
	next if (-d $dir);
	message(3, "Making dir $dir\n");
	eval { io($dir)->mkpath };
	if ($@) {
	    message(1, "ERROR - Could not create directory $dir: $@\n");
	    confess $Yaxx::FATAL_ERROR;
	}
    }

    1;
}

##****************************************************************************
sub fakedir {
##****************************************************************************
    my $self = shift;

    # If fakeid evals to true (i.e. fakeid is defined and not 0, either of which
    # would imply this is the real dataset) then return a sub-dir "fakeid/"
    return $self->fakeid ? "fake" . $self->fakeid . '/' : '';
}

##****************************************************************************
sub clean_files {
##****************************************************************************
    my $self = shift;
    my $par  = shift;
    my $type_list = $par->{files};
    local $_;
    return 1 unless $type_list;

    my $src_dir = io($self->file('src_dir'))->chdir;
    my $clean_spec = $self->interpolate_struct( $self->clean_spec );
    
    # Find all the files associated with spectral fitting
    my @fit_files;
    foreach my $mdl (map {basename $_,'.in'} glob("*.in")) {
	push @fit_files, map {"$mdl.$_"} listify($clean_spec->{fit_model}{ext});
    }

    # Define how to expand the groups, either in terms of base types
    # or by a list of files from a glob
    my %clean_types = %{$clean_spec->{types}};
    my %clean_files = (fit => [@fit_files]);  # Special case of fit products
    while (my ($key, $glob) = each %{$clean_spec->{globs}}) {
	$clean_files{$key} = [ glob($glob) ];
    }

    my @clean;
    my @type = split ' ', $type_list;
    while (my $type = shift @type) {
	# Try to:
	#   1) expand the type into component types
	#   2) find a list of files for that type
	#   3) find the single file for that type
	if ($clean_types{$type}) {
	    push @type, @{$clean_types{$type}};
	} elsif ($clean_files{$type}) {
	    push @clean, @{$clean_files{$type}} if @{$clean_files{$type}};
	} else {
	    my $file;
	    eval { $file = $self->file($type) };
	    if ($@) {		
		message(1, "WARNING - Unknown file type '$type' for cleaning\n");
	    } else {
		push @clean, $file if defined $file;
	    }
	}
    }

    if (@clean) {
	message(3, "Cleaning files:\n   " . join("\n   ",@clean) . "\n");
	map { unlink $_ if (-e $_ or -l $_) } @clean;
    }

    1;
}

##****************************************************************************
sub check_for_pause {
#
# If 'pause_file' exists in either current or home directory, then wait until
# it goes away    
##****************************************************************************
    my $self = shift;
    my $waiting;
    while (-e $self->file('pause') or -e "$ENV{HOME}/" . $self->file('pause')) {
	unless ($waiting) {
	    message(1, "Pause file detected.  Waiting until it disappears...\n", time=>1);
	    $waiting = 1;
	}
	sleep 30;
    }
    return 1;
}
	
##****************************************************************************
sub get_lock {
##****************************************************************************
    my $self = shift;
    my $msg;

    return 1 unless $self->lock;

    my $lock_file = $self->file('lock');
    my $date = gmtime();

    if (-r $lock_file) {
	$msg < io($lock_file);
	message(2, $msg);
	return;
    } else {
	$msg = "Source locked by $ENV{USER} at $date by process $PROCESS_ID on $ENV{HOST}\n";
	write_file($lock_file, $msg);
    }

    return 1;
}

##****************************************************************************
sub release_lock {
##****************************************************************************
    my $self = shift;

    my $lock_file = $self->file('lock');
    return ((-w $lock_file and unlink $lock_file) or not $self->lock);
}

##****************************************************************************
sub file {
#
# Encapsulate all the naming conventions here, using the file_definition
# configuration variable
#
##****************************************************************************
    my ($self, $arg, $opt) = @_;
    my ($type, @ext) = split /\./, $arg;

    my $f = $self->file_definition->{$type};
    if (not defined $f) {
	message(1, "ERROR - invalid file type '$type' detected\n");
	confess $Yaxx::FATAL_ERROR;
    }

    # Strip any whitespace which is inserted to improve readability
    $f =~ s/\s//g;

    # Temporarily override specified yaxx variables.  The file() options ABSOLUTE
    # and CONTENT are upper-cased to hopefully avoid collisions with user source data columns
    ($self->can($_) ? local $self->{$_} : local $self->{source_data}->{$_}) = $opt->{$_}
      for keys %{$opt};

    eval { $f = $self->interpolate_string($f) };
    if ($@) {
	# Can't recover from an error interpolating string here
	message(1, "ERROR in Yaxx::file() - $@\n");
	confess $Yaxx::FATAL_ERROR;
    }

    $f .= join('.', '', @ext) if @ext;

    # Return the full file path if absolute, else prepend the output_dir
    # and return relative path
    my $file = io($f)->is_absolute ? io($f) : io->join($self->output_dir, $f);

    # WTF?
    $file->name;
    my $filepath = $opt->{ABSOLUTE} ? $file->rel2abs : $file->abs2rel;
    # $file->relative/absolute->pathname doesn't work if file doesn't exist!
    $filepath = '.' unless $filepath =~ /\S/;

    # If a file prefix is 
    if (defined $opt->{PREFIX} or defined $opt->{SUFFIX}) {
	my @path = io($filepath)->splitdir;
	$path[-1] = $opt->{PREFIX} . $path[-1] if defined $opt->{PREFIX};
	$path[-1] = $path[-1] . $opt->{SUFFIX} if defined $opt->{SUFFIX};
	$filepath = io->join(@path);
    }

    # If file content is requested (instead of file name), grab content and return it
    if ($opt->{CONTENT}) {
	my $file_content;
	eval { $file_content < io($filepath) };
	confess "ERROR - Failed to read file contents for $arg file ($filepath): $@\n" if ($@);
	return $file_content;
    }

    return $filepath;
}

########################################################################################
sub get_resource_files {
# Make sure that all yaxx resource files are available in the analysis directory
########################################################################################
    my $self = shift;

    my $resources_dir = io($self->file('resources_dir'))->mkpath->absolute->pathname;

    # Go to resources_yaxx_dir, get files recursively and copy structure to $resources_dir
    my $resources_yaxx_dir = io($self->file('resources_yaxx_dir'))->chdir;

    for my $yaxx_file (io('.')->deep->all ) {
	my $local_file = io->join($resources_dir, $yaxx_file->relative->pathname); #  . '/' .
	next if grep(/^\.svn$/, $yaxx_file->splitdir);
	if ($yaxx_file->is_dir) {
	    io($local_file->name)->mkpath;
	} else {
	    # Copy unless local file exists and local file is newer than yaxx file
	    $yaxx_file > $local_file unless ($local_file->exists
					     and $local_file->mtime > $yaxx_file->mtime );
	}
    }

    return 1;
}

########################################################################################
sub get_data_files {
########################################################################################
    my $self = shift;
    my $par  = { min_files => 1,
		 %{ shift() }
	       };

    my @out_files = listify($par->{out_file});

    # Go through globs looking for matching files.
    # Use funky bsd_glob to ensure correct behavior if the glob doesn't contain wildcards.
    # The second argument overrides the default flags that make glob return the pattern
    # if it does not contain a wildcard, even if that file doesn't exist.  Why???
    my $glob = $par->{file_glob};
    my @files;
    foreach (split ' ', $glob) {
	push @files, bsd_glob($_, undef);
    }

    # Only accept files with read permission
    @files = grep { -r } @files;

    # Check that specified number of files were found
    if (   (@files < $par->{min_files})
	or (defined $par->{max_files} and @files > $par->{max_files})) {
	my $err = @files ? 'too many' : 'no';
	message(1, "ERROR - $err data files matching $glob found for "
		. $self->unique_identifier . "\n");
	foreach (@files) {
	    message(1, "     $_\n");
	}
	return;
    }

    # Make a local unzipped copy of the event file in the yaxx directory.  If the
    # event file in the database is already unzipped, then just link to it
    # Return an absolute (full) filename

    if (@files == 1) {
	make_local_copy($files[0], $out_files[0], $par->{copy} ? {copy=>1} : () ); 
    } elsif (@files > 1) {
        # Run dmmerge.  
	# Delete existing file (to force rebuild) if it is older than any of the input files

	if (test_dep(-target => $par->{out_file}, -depend => \@files)) {
	    # Multiple files, need to merge them
	    my $list = join ',', @files;
	    my $ok = run_tool("dmmerge",
			      infile => $list,
			      columnList => '',
			      outfile => $out_files[0],
			      outBlock => '',
			      clobber => 'yes',
			      lookupTab => "$ENV{ASCDS_CALIB}/dmmerge_header_lookup.txt",
			      { timeout => $self->timeout,
				paste => 1,
				loud => 1,
			      }
			     );
	    return unless $ok;
	}
    }

    # Make any additional local copies
    make_local_copy($out_files[0], $out_files[$_], $par->{copy} ? {copy=>1} : ())
      for (1..$#out_files);

    return 1;
}
    
#********************************************************************************
sub store_object {
# Store a copy of the object (to a depth of 1) in the object_dump file
# in the source directory
#********************************************************************************
    my $self = shift;
    
    local $Data::Dumper::Maxdepth = 5;
    local $Data::Dumper::Sortkeys = 1;

    my %save;
    my @ignore = qw(next_source prev_source);
    foreach (@ignore) {
	$save{$_} = $self->{$_};
	delete $self->{$_};
    }

    my $dumper = Dumper $self;
    eval { $dumper > io($self->file('object_dump')) };
    if ($@) {
	message(1, "ERROR - $@\n");
	confess $Yaxx::FATAL_ERROR;
    }

    foreach (@ignore) {
	$self->{$_} = $save{$_};
    }

    return 1;
}

#********************************************************************************
sub restore_object {
# Restore the stored object and return it
#********************************************************************************
    my $self = shift;
    
    my $dumper_output;
    my $dumper_file = $self->file('object_dump');
    eval { $dumper_output < io($dumper_file) };
    if ($@) {
	message(1, "WARNING - unable to find stored information file for ObsId "
		. $self->{obsid} . ", Src " . $self->{srcid}
		. ".  Was it processed?\n");
	return;
    }

    # Eval the expression in the file, which sets a variable that gets
    # returned as the output of eval
    my $VAR1;			# The var defined by Dumper output
    eval "$dumper_output";
    if ($@) {
	message(1, "WARNING - unable to parse stored information file for ObsId "
		. $self->{obsid} . ", Src " . $self->{srcid}
		. "\n");
	message(1, "$@\n");
	return;
    }
    
    return $VAR1;
}    

#********************************************************************************
sub open_log {
#
# Open a log file (using 'message') for processing related to this source
#********************************************************************************
    my $self = shift;
    my $log = $self->file('log');

    if (-e $log) {
	my $max = 1;
	map { $max = $1+1 if (/${log}\.(\d+)/ and $1 >= $max) } glob("${log}*");
	move($log, "$log.$max");
    }

    # Create a message object to handle source output
    my $msg = Ska::Message->new(stdout  => undef,
				file    => $log,
				auto_cr => undef,
			       );
    # Add message object to Yaxx::message routine
    message($msg, 0, $self->verbose_source_log);

    # Maintain a reference to the Ska::Message object for closing later
    $self->{ska_msg_obj} = $msg;
    1;
}

#********************************************************************************
sub close_log {
#********************************************************************************
    my $self = shift;
    message($self->{ska_msg_obj}, { close => 1 });
}


#********************************************************************************
sub get_obs_info_from_pi {
#********************************************************************************
    my $self = shift;
    my $ok;

    return unless (-e $self->file('pi') and -e $self->file('bg_pi'));

    my %rmf = fits_read_bintbl($self->file('rmf'), {extname => 'EBOUNDS'});
    $ok = PDL::which($rmf{e_min} > $self->min_energy_kev);
    my $min_bin = ($ok->nelem > 0) ? $ok->at(0)+1 : 1;

    $ok = PDL::which($rmf{e_max} < $self->max_energy_kev);
    my $max_bin = ($ok->nelem > 0) ? $ok->at(-1)+1 : $rmf{e_max}->nelem;

    my %counts;
    for my $type (qw(pi bg_pi)) {
	run_tool('dmstat',
		 infile => $self->file($type) . qq{'[channel=${min_bin}:${max_bin}][cols counts]'},
		 { loud => 1,
		 },
		);
	
	run_tool('pget dmstat out_sum',
		 { out => \$counts{$type},
		   loud => 1,
		 }
		);
	$counts{$type} =~ s/\s//g;
    }

    my $hdr = fits_read_hdr($self->file('pi'), 'spectrum');
    my $hdr_bkg = fits_read_hdr($self->file('bg_pi'), 'spectrum');
    $self->counts( $counts{pi} - $counts{bg_pi} * $hdr->{BACKSCAL} / $hdr_bkg->{BACKSCAL} );
    $self->exposure( $hdr->{EXPOSURE} );

    return 1;
}

##****************************************************************************
sub make_local_copy {
#
#  Usage:
#    make_local_copy(in_file, [out_file], [{copy=>1, absolute=>1}])
#    
#  Make a local copy of or link to the specified in_file.  Gunzip
#  it if necessary.  If out_file is not supplied, then use the
#  basename of the in_file.  Don't do anything if the output file is 
#  already there.  Finally, if the 'copy' option is specified in the
#  optional hash reference (last argument), then make an actual copy
#  instead of just linking (if the in_file is not gzipped).
#
##****************************************************************************
    return unless @_;		# Need at least 1 arg
    my $cwd = cwd;

    # First pull any hash ref off the end 
    my %par = (freshen => 1);   # Freshen an outdated local copy
    if (ref $_[-1] eq 'HASH') {
	%par = %{pop @_};
    }

    my $in_file = shift(@_) or return;
    return unless (-e $in_file);
    
    my $out_file;
    if (@_) {
	# Explicit file name spec'd
	$out_file = shift;
    } else {
	# Use input file name, and assume it is going in the current directory
	$out_file = basename($in_file);
	$out_file =~ s/\.gz$//;	# Remove any trailing .gz
    }

    $par{absolute} ||= $in_file =~ /^\//;

    # Define absolute file name for $in_file and $out_file
    $in_file = File::Spec->rel2abs($in_file, $cwd);

    unless ($par{absolute}) {
	# Go to the directory spec'd for $out_file
	chdir dirname($out_file) 
	  or croak "ERROR - could not change dir to make local copy of $out_file\n";
	$out_file = basename $out_file;
	# Now make in_file relative to the out_file directory
	$in_file = File::Spec->abs2rel($in_file);
    } else {
	$out_file = File::Spec->rel2abs($out_file, $cwd);
    }

    # Delete outfile if it exists and is older than input file, and 'freshen' par allows
    unlink $out_file if (-e $out_file and ($par{force}
					   or ($par{freshen}
					       and [stat $in_file]->[9] > [stat $out_file]->[9])));

    # Now gunzip or link, unless out_file exists and is older than the input file
    unless (-e $out_file) {
	my $copy_cmd = exists $par{copy} ? "cp" : "ln -s";
	system( $in_file =~ /\.gz$/ ?
		"gunzip --stdout $in_file > $out_file" : 
		"$copy_cmd $in_file $out_file");
	message(3, "Making local file " . File::Spec->rel2abs($out_file) . "\n");
    }

    chdir $cwd;
    return $out_file;
}

1;
