# --8<--8<--8<--8<--
#
# Copyright (C) 2006 Smithsonian Astrophysical Observatory
#
# This file is part of yaxx
#
# yaxx is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# yaxx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the 
#       Free Software Foundation, Inc. 
#       51 Franklin Street, Fifth Floor
#       Boston, MA  02110-1301, USA
#
# -->8-->8-->8-->8--

package Yaxx;
use strict;
use warnings;

##****************************************************************************
sub make_latex_report {
#
# Make latex/postscript report that summarizes source and fit results.  
#
##****************************************************************************
    my $self = shift;
    
    #  Go to the source directory and find some files for dependency check
    chdir dirname $self->file('src_dir');
    my $report_template_file = $self->file('report_template.tex');
    my $report_file          = $self->file('report');
    my $report_tex_file      = $self->file('report.tex');
    my $report_ps_file       = $self->file('report.ps');
    my @mdl_files            = glob("*.mdl");

    my @target = ( $report_ps_file );
    my @test_dep = (-target => \@target, -depend => [@mdl_files, $report_template_file]);
    return 1 unless test_dep(@test_dep);

    message(3, "Making LaTeX report\n");

    # Clean up
    map { unlink "$report_file.$_" if -e "$report_file.$_" } qw (ps tex aux dvi);

    # Get the structure of fit results
    my $fte = $self->formatted_table_elements('latex');

    # Make the latex for the actual fit parameter elements in the fit table
    my $param_table;
    for my $i (0..$#{$fte->{source_model_names}}) {
	$param_table .=  latex_verbatim($fte->{source_model_names}[$i]) . " & ";
	$param_table .=  join(" & ", @{$fte->{param_values_table}[$i]});
	$param_table .=  "\\\\ \n";
    }

    # Make latex for the spectral fit plots
    my $count = 1;
    my $fit_plots;
    my $width_inches = $self->{fit_plot}{width_inches};
    for my $source_model (@{$fte->{source_model_names}}) {
	$fit_plots .= "\\resizebox{${width_inches}in}{!}{\\includegraphics{${source_model}.ps}} \n";
	$fit_plots .= "\\vspace{0.1in} \\\\ \n" if ($count++ % $self->{fit_plot}{n_across} == 0);
    }

    # Define the special substitution vars for the latex report_template file
    my @subst = (COLUMN_ALIGN    => 'l' . 'c' x @{$fte->{columns}},
		 COLUMN_HEADERS1 => join(" & ", @{$fte->{column_headers1}}),
		 COLUMN_HEADERS2 => join(" & ", @{$fte->{column_headers2}}),
		 COLUMN_UNITS    => join(" & ", @{$fte->{units}}),
		 PARAM_VALUES    => $param_table,
		 FIT_PLOTS       => $fit_plots,
		);

    # Make the actual report.  Read in template then interpolate accordingly
    my $report < io($report_template_file);
    $report = $self->interpolate_string($report,
					@subst,
				       );
    
    eval { $report > io($report_tex_file) };
    if ($@) {
	message(1, "ERROR - Could not open LaTeX file '$report_tex_file' for writing: $@\n");
	return;
    };

    run_tool("latex -interaction=batchmode $report_tex_file", {timeout => 120, loud => 1});
    unless (-e "$report_file.dvi") {
	message(1, "ERROR - latex failed to run correctly\n");
	return;
    }

    run_tool("dvips $report_file -o", {timeout => 120, loud => 1});
    unless (-e "$report_file.ps") {
	message(1, "ERROR - dvips failed to run correctly\n");
	return;
    }

    map {unlink "$report_file.$_"} qw(aux dvi log);
    1;
}

##****************************************************************************
sub make_html_report {
#
# Make HTML report that summarizes source and fit results.  
#
##****************************************************************************
    my $self = shift;
    my $fit_plots = '';		# Fit plots
    my $fte_html = '';		# Formatted (HTML) fit table elements 
    my $process_failure_msg = '';
    
    # No report for fake datasets
    # return 1 if $self->fakeid; 
    
    message(3, "Making HTML report\n");

    # Make the HTML to link to prev, next reports and master index
    my $index_button = q{<a href="} . $self->file('report_index.html') . 
                       q{"> <img src="} . $self->file('index_button') . q{" alt="Index" border=1></a>};
    my $prev_button  = defined $self->prev_source ? 
                         q{<a href="} . $self->prev_source->file('report.html') . 
                         q{"> <img src="} . $self->file('prev_button')  . q{" alt="Previous" border=1></a>}
			   : q{<img src="} . $self->file('empty_button') . q{">};
    my $next_button  = defined $self->next_source ? 
                         q{<a href="} . $self->next_source->file('report.html') . 
                         q{"> <img src="} . $self->file('next_button') . q{" alt="Next" border=1></a>}
			   : q{<img src="} . $self->file('empty_button') . q{">};

    # If processing was successful then generate HTML for fit values table and fit plots
    if (defined $self->failed_at) {
	$process_failure_msg = "ERROR - processing failed at step " . $self->failed_at;
    } else {
	# Get the fit parameter data (properly formatted for HTML) and put into a table
	my $fte = $self->formatted_table_elements('html');
	$fte_html = $self->html_fit_table($fte);

	my $count = 1;
	for my $source_model (@{$fte->{source_model_names}}) {
	    $fit_plots .= qq{<img src="${source_model}.jpg">\n};
	    $fit_plots .= qq{<br />\n} if ($count++ % $self->{fit_plot}{n_across} == 0);
	}
    }

    # Make the actual report.  Read in appropriate template then interpolate accordingly
    my $report < io($self->file('report_template.html'));
    $report = $self->interpolate_string($report,
					FIT_TABLE 	=> $fte_html,
					FIT_PLOTS 	=> $fit_plots,
					LOG_FILES_TABLE => $self->html_log_files_table,
					PREVIOUS_REPORT => $prev_button,
					NEXT_REPORT     => $next_button,
					INDEX           => $index_button,
					PROCESS_FAILURE => $process_failure_msg,
				       );
    
    eval { $report > io($self->file('report.html')) };
    if ($@) {
	message(1, "ERROR - Could not open HTML file for writing: $@\n");
	return;
    };

    1;
}


##****************************************************************************
sub html_log_files_table {
##****************************************************************************
    my $self = shift;
    my @log_files = grep { $_->filename =~ /\A log (\.\d+)? \Z/x } io($self->file('src_dir'))->all;
    @log_files = sort { $b->mtime <=> $a->mtime } @log_files;
    my @table_data;
    push @table_data, [qw(File Date Size)];
    for (@log_files) {
	push @table_data, [ q{<a href="} . $_->filename . q{">} . $_->filename . q{</a>},
			    scalar localtime($_->mtime),
			    $_->size,
			  ];
    }

    my $html_table = HTML::Table->new(-class => "style1",
				      -data => \@table_data,
				     );
    $html_table->setRowHead(1);

    return $html_table->getTable;
}
##****************************************************************************
sub html_fit_table {
##****************************************************************************
    my $self = shift;
    my $fte = shift;
    my @table;
    push @table, [ '', @{$fte->{column_headers1}} ];
    push @table, [ '', @{$fte->{column_headers2}} ];
    push @table, [ '', @{$fte->{units}} ];

    my @param_table;
    for (0..$#{$fte->{source_model_names}}) {
	push @table, [ $fte->{source_model_names}[$_], @{$fte->{param_values_table}[$_]} ];
    }

    my $table = HTML::Table->new(-class => "style1",
				 -data  => \@table,
				);
    $table->setRowHead(1);
    $table->setRowHead(2);
    $table->setRowHead(3);
    $table->setColHead(1);

#    # Set the formatting of columns
#    for (0..$#col) {
#	my $ec = $opt{event_table}{$col[$_]};
#	$table->setColAlign($_+1, $ec->{align}) if $ec->{align};
#	my $pre = $ec->{format_pre} || '';
#	my $post = $ec->{format_post} || '';
#	$table->setColFormat($_+1, $pre, $post);
#    }
#
#    # Set the color of each row
#    for (0..$#row_bg_color) {
#	if (defined $row_bg_color[$_]) {
#	    my $bg_color = $opt{color}{$row_bg_color[$_]};
#	    $table->setRowBGColor($_+2, "#$bg_color") if defined $bg_color;
#	}
#	if (defined $row_font_color[$_]) {
#	    my $font_color = $opt{color}{$row_font_color[$_]};
#	    $table->setRowStyle($_+2, "color:#$font_color") if defined $font_color;
#	}
#    }
#
#    $table->setRowAlign(1, 'CENTER');
#    $table->setCaption("<span style=$opt{web_page}{table_caption_style}>Chandra Events</span>", 'TOP');
#
    return $table->getTable;
}

#********************************************************************************
sub format_var {
#
#  Format a yaxx variable for output.  
#********************************************************************************
    my $self = shift;
    my $var  = shift;
    my $opt  = shift || {};
    my $language = $opt->{type} || 'plain';# Type of output (plain (default), html, latex)
				# Used to encode special characters for this system
    
    my $sprintf_format = $opt->{DEFAULT_FMT} ? $self->get_formatting_rule($var)->{fmt}
                                             : ($opt->{FORMAT} || '%s');
    
    # Get rid of any leading/trailing spaces
    $var =~ s/\A \s+ | \s+ \Z//xg;

    # Try the class accessor for $var first.  If this fails, use 'source_data', which
    # is the data about the object supplied by user in the objlist or detect file
    my $val = $self->get_var($var); 

    # Format defined variable or else return the 'undef_value' (i.e. what to
    # use if value is undef) from the formatting_rules definitions. 
    $val = defined $val ? sprintf($sprintf_format, $val) : $self->get_formatting_rule($var)->{default};

    # Make string 'safe' (encode special chars) for the given output type
    if ($language eq 'html') {
	$val = HTML::Entities::encode($val);
    } elsif ($language eq 'latex') {
	$val = latex_verbatim($val);
    }

    return $val;
}

#********************************************************************************
sub fit_table_from_MDL_files {
# Parse the MDL files for given model fits and return in a nice structure
# In: list of MDL files
# Out: Structure with parameter fit values etc
#********************************************************************************
    my $self = shift;
    my $MDL_file = shift;	# Optional MDL file name

    my $mdl_dir = $self->file('src_dir') . "/*.mdl";
    my @MDL_files = $MDL_file || glob($mdl_dir);

    my %table;
    my ($value, $upper, $lower);

    foreach $MDL_file (@MDL_files) {
	(my $source_model_name = basename $MDL_file) =~ s/\.mdl \Z//x;
	my %mdl = fits_read_bintbl("${MDL_file}\[MDL_Models\]");
	my $mdl_header = fits_read_hdr($MDL_file, "MDL_Models");
	push @{$table{source_model_names}}, $source_model_name;
	for my $i (0..$#{$mdl{model}}) {
	    my $parname = $mdl{parname}->[$i];

	    if (defined $parname and $parname =~ /\S/) {
		$value = $mdl{parvalue}->at($i);
		$upper = exists $mdl{proj_upper} ? $mdl{proj_upper}->at($i)
		  : (exists $mdl{unc_upper} ? $mdl{unc_upper}->at($i) : 0);
		$lower = exists $mdl{proj_lower} ? $mdl{proj_lower}->at($i)
		  : (exists $mdl{unc_lower} ? $mdl{unc_lower}->at($i) : 0);
		# Keep track of columns for table in order
		push @{$table{param_names}}, $parname;
		$table{fit_values}{$source_model_name}{$parname} = {value => $value,
								    lower => $lower,
								    upper => $upper,
								   };
	    }
	    if ($mdl{statname}->[$i] =~ /\S/) {
		$table{fit_statistic}{$source_model_name} = { name  => $mdl{statname}->[$i],
							      value => $mdl{statval}->at($i),
						       dof   => $mdl_header->{F_DOF},
						     };
	    }
	}
    }
    # Keep only unique fit parameter names
    @{$table{param_names}} = List::MoreUtils::uniq @{$table{param_names}};

    return \%table;
}

#********************************************************************************
sub formatted_table_elements {
#
# Go through fit table (from MDL files) and massage into nicely formatted
# column elements (e.g. values with uncertainties) for output in a report.
# The language of the report (html or latex) is an input.
#********************************************************************************
    my $self = shift;
    my $language = shift;

    # Parse the MDL files and get the perl structure representing all the fit parameters
    # with uncertainties    
    my $fit_table = $self->fit_table_from_MDL_files();

    # Go through table and do necessary formatting and cleanup for a formatted table
    # Get together info for column headers
    my @columns;
    my @column_headers1;
    my @column_headers2;
    my @units;
    my %format;
    foreach my $parname (@{$fit_table->{param_names}}) {
	$format{$parname} = $self->get_formatting_rule($parname);
	next if $format{$parname}->{ignore};
	push @columns, $parname;

	if ($parname =~ /(\S+)\.(\S*)/) {
	    push @column_headers1, $1;
	    push @column_headers2, $2;
	} else {
	    message(1, "WARNING - Could not parse parameter name '$parname'\n");
	    next;
	}
	push @units, $format{$parname}->{"unit_$language"};
    }

    # Add column for the fit statistic
    push @columns, 'fit_statistic';
    push @column_headers1, 'Fit';
    push @column_headers2, 'stat';
    push @units, '(DOF)';

    # Fill out the parameter values table, consisting of one row for each source model
    # and the parameter columns defined above
    my @source_model_names = @{$fit_table->{source_model_names}};
    my @param_values_table;
    foreach my $source_model (@source_model_names) {
	my @param_values_row;
	foreach my $parname (@columns) {
	    my $table_entry;
	    # If this param exists in currennt source_model, then format and include in table
	    if (defined (my $fit_record = $fit_table->{fit_values}{$source_model}{$parname})) {
		my $value = $fit_record->{value};
		my $upper = $fit_record->{upper};
		my $lower = $fit_record->{lower};

		$lower = $self->fix_uncertainty($language, 'report', $parname, $value, $lower, 'lower');
		$upper = $self->fix_uncertainty($language, 'report', $parname, $value, $upper, 'upper');
		$table_entry = format_value_with_limits($language,
							$format{$parname},
							$value, $lower, $upper);
	    } elsif ($parname eq 'fit_statistic') {
		# Add special entry for fit statistic.  (Should not be hard coded.  Fix this)
		$table_entry = sprintf("%.1f (%d)",
				       $fit_table->{fit_statistic}{$source_model}{value},
				       $fit_table->{fit_statistic}{$source_model}{dof});
	    } else {
		$table_entry = '';
	    }
	    push @param_values_row, $table_entry;
	}

	push @param_values_table, [ @param_values_row ];
    }
	
    return { columns => \@columns,
	     column_headers1 => \@column_headers1,
	     column_headers2 => \@column_headers2,
	     param_values_table => \@param_values_table,
	     units => \@units,
	     source_model_names => \@source_model_names,
	   };
}

#********************************************************************************
sub summary {
#********************************************************************************
    my $self = shift;
    my $mdl_name = shift;
    local $_;

    return if (-e $self->file('ignore_src') or -e $self->file('ignore_obs'));

    my %table;
    my @cols;
    my %info;
    my @info_cols = listify($self->summary_header_key_col);

    my ($value, $upper, $lower);

    # Try to find mdl file
    my $file = $self->file('src_dir') . "/${mdl_name}.mdl";
    return unless -e $file;	# Return (without issuing a warning) if no file found.
				# This likely means this model is not fit for this source

    my $fit_table = $self->fit_table_from_MDL_files($file);

#    my %mdl = fits_read_bintbl("$file\[MDL_Models\]");
    my $mdl_header = fits_read_hdr("$file", "MDL_Models");

    foreach ( qw(obsid ccdid srcid), $self->fake_instances ? 'fakeid' : () ) {
	$table{$_} = $self->{$_};
	push @cols, $_ if defined $table{$_};
    }

    # Insert supplemental information from header if possible
    message(1, "WARNING - No supplemental summary info in $file\n") unless (defined $mdl_header->{RA});
    foreach (@info_cols) {
	if (defined $mdl_header->{$_}) {
	    $table{lc $_} = $mdl_header->{$_};
	    push @cols, lc $_;
	}
    }
    
    # OK, go through param names, see if it is in the current model under consideration.
    # If so, put in the @cols and the corresponding data table.
    foreach my $param_name (@{$fit_table->{param_names}}) {
	push @cols, $param_name, "${param_name}.up", "${param_name}.low";
	$table{"${param_name}"}     = $fit_table->{fit_values}{$mdl_name}{$param_name}{value};
	$table{"${param_name}.up"}  = $fit_table->{fit_values}{$mdl_name}{$param_name}{upper};
	$table{"${param_name}.low"} = $fit_table->{fit_values}{$mdl_name}{$param_name}{lower};
    }

    # Add all of the column values for this source from the user-supplied object list file (but
    # not if a value of the same name has already been added).  These values could be redshift,
    # optical mags, classification, etc etc.  
    foreach (keys %{$self->source_data}) {
	unless (defined ($table{$_})) {
	    push @cols, $_;
	    $table{$_} = $self->source_data->{$_};
	}
    }

    return { cols => [ @cols ],
	     data => { %table },
	   };
}

#********************************************************************************
sub get_formatting_rule {
# 
# Find the model format name glob that matches the param name
#********************************************************************************
    my $self = shift;
    my $parname = shift;
    local $_;

    return unless ($parname =~ /\S/);

    # Get the hash ref with all the formatting rules
    my $mf = $self->formatting_rule;

    # Get the length of the non-wildcard characters in each rule name
    my %len;
    foreach (keys %{$mf}) {
	(my $no_wildcard = $_) =~ s/[\*\?]//g;
	$len{$_} = length($no_wildcard);
    }
    
    # Sort the rules based on the length of the non-wildcard chars in the rule name.
    # This ensures that '*' would be first and before e.g. 'X'
    my @rules_by_length = sort { $len{$a} <=> $len{$b} } keys %{$mf};

    # Go through rules (in order) and push matches onto output hash %rule_out.
    # The matching is not case sensitive.
    my %rule_out = ();
    foreach my $rule_name (@rules_by_length) {
	if (Text::Glob::match_glob(lc($rule_name), lc($parname))) {
	    %rule_out = (%rule_out,
			 %{$mf->{$rule_name}}
			);
	}
    }

    return \%rule_out;
}

#********************************************************************************
sub fix_uncertainty {
# Adjust uncertainty values (upper,lower limits) based on uncertainty rules.
# NaN values in MDL file indicate uncertainty pegged
#********************************************************************************
    my $self = shift;
    my ($language,		# 'plain', 'html', or 'latex'
	$type,			# 'report' or 'summary'
	$parname,		# Parameter name (e.g. abs.nh)
	$parval,		# Paramter value
	$limit_in,		# Input lower or upper uncertainty limit
	$limit_dir		# 'lower' or 'upper'
       ) = @_;
    my $limit_out;

    # Check if the lower/upper limit is set to NaN, which is the value assigned
    # in parse_limits_table (i.e. parsing output of projection) if limit is not found.
    # In that case we try to put in a nicer value for reports/summaries
    return $limit_in unless ($limit_in != $limit_in);

    my $mf = $self->get_formatting_rule($parname);

    $limit_out = $mf->{"${type}_${limit_dir}"};

    my %inf = (plain => 'Inf',
	       html  => '&#8734;',
	       latex => '\infty',
	       );
    
    return $self->interpolate_string($limit_out,
				     INFINITY => $inf{$language},
				     VALUE => $parval,
				    );
}

#********************************************************************************
sub latex_verbatim {
#
# Fix any characters which have special meaning to latex, e.g. _ => \_
#********************************************************************************
    local $_;
    $_ = shift;
    my @fix = ( qr{\\} => '\$\\backslash\$',
		qr{%} => '\\%',
		qr{\$} => '\\\$',
		qr{<} => '\$<\$',
		qr{>} => '\$>\$',
		qr/}/ => '\$\\}\$',
		qr/{/ => '\$\\{\$',
		qr/&/ => '\\&',
		qr/#/ => '\\#',
		qr/\^/ => '\\^',
		qr/_/ => '\\_',
		qr/~/ => '\\~{}',
		);

    # Go through the fixes (in order!) and apply globally
    my ($plain, $tex);
    while ($plain = shift @fix and $tex = shift @fix) {
	s/$plain/$tex/g;
    }

    return $_;
}

#********************************************************************************
sub format_value_with_limits {
#********************************************************************************
    my ($language,
	$formatting_rule,
	$value,
	$lower,
	$upper
       ) = @_;
    my $mult = $formatting_rule->{mult};
    my $fmt = $formatting_rule->{fmt};
    my $fmt_value = sprintf $fmt, $value * $mult;
    my $fmt_value_limits;

    return $fmt_value unless ($lower or $upper);

    # Format the lower/upper limits if defined
    for ($lower, $upper) {
	if ($_) {  # Hopefully works to detect that limit is defined and not zero
	    $_ = sprintf("$fmt", abs($_*$mult)) if /\A $RE{num}{real} \Z/x;
	} else {
	    $_ = q{};
	}
    }

    if ($language eq 'latex') {
	$fmt_value_limits = $fmt_value . '$' . "_{-$lower}^{+$upper}" . '$';
    } elsif ($language eq 'html') {
	my $unc_table = new HTML::Table(-class => "uncertainty",
					-data  => [["+$upper"],["-$lower"]],
				   );
	
	my $table = new HTML::Table(-data  => [[ $fmt_value, $unc_table->getTable ]],
				   );

       $fmt_value_limits = $table->getTable;
    }

    return $fmt_value_limits;
}

1;
